import numpy as np
from pathlib import Path
import pandas as pd
import pandas.testing as tm
import pytest
from tableauhyperapi import Connection, TableName, HyperProcess, Telemetry, CreateMode

import pantab
import pantab._compat as compat


def assert_roundtrip_equal(result, expected):
    """Compat helper for comparing round-tripped results."""

    expected["float32"] = expected["float32"].astype(np.float64)

    if compat.PANDAS_100:
        expected["object"] = expected["object"].astype("string")
        expected["non-ascii"] = expected["non-ascii"].astype("string")

    tm.assert_frame_equal(result, expected)


def test_basic(df, tmp_hyper, table_name, table_mode):
    # Write twice; depending on mode this should either overwrite or duplicate entries
    pantab.frame_to_hyper(df, tmp_hyper, table=table_name, table_mode=table_mode)
    pantab.frame_to_hyper(df, tmp_hyper, table=table_name, table_mode=table_mode)
    result = pantab.frame_from_hyper(tmp_hyper, table=table_name)

    expected = df.copy()

    if table_mode == "a":
        expected = pd.concat([expected, expected]).reset_index(drop=True)

    assert_roundtrip_equal(result, expected)


def test_multiple_tables(df, tmp_hyper, table_name, table_mode):
    # Write twice; depending on mode this should either overwrite or duplicate entries
    pantab.frames_to_hyper(
        {table_name: df, "table2": df}, tmp_hyper, table_mode=table_mode
    )
    pantab.frames_to_hyper(
        {table_name: df, "table2": df}, tmp_hyper, table_mode=table_mode
    )
    result = pantab.frames_from_hyper(tmp_hyper)

    expected = df.copy()
    if table_mode == "a":
        expected = pd.concat([expected, expected]).reset_index(drop=True)

    # some test trickery here
    if not isinstance(table_name, TableName) or table_name.schema_name is None:
        table_name = TableName("public", table_name)

    assert set(result.keys()) == set((table_name, TableName("public", "table2")))
    for val in result.values():
        assert_roundtrip_equal(val, expected)


def test_roundtrip_with_external_hyper_process(df, tmp_hyper):
    default_log_path = Path.cwd() / "hyperd.log"
    if default_log_path.exists():
        default_log_path.unlink()

    # By passing in a pre-spawned HyperProcess, one can e.g. avoid creating a log file
    parameters = {"log_config": ""}
    with HyperProcess(
        Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU, parameters=parameters
    ) as hyper:
        # test frame_to_hyper/frame_from_hyper
        pantab.frame_to_hyper(df, tmp_hyper, table="test", hyper_process=hyper)
        result = pantab.frame_from_hyper(tmp_hyper, table="test", hyper_process=hyper)
        assert_roundtrip_equal(result, df)

        # test frame_from_hyper_query
        result = pantab.frame_from_hyper_query(
            tmp_hyper, "SELECT * FROM test", hyper_process=hyper
        )
        assert result.size == 63

        # test frames_to_hyper/frames_from_hyper
        pantab.frames_to_hyper(
            {"test2": df, "test": df}, tmp_hyper, hyper_process=hyper
        )
        result = pantab.frames_from_hyper(tmp_hyper, hyper_process=hyper)
        assert set(result.keys()) == set(
            (TableName("public", "test"), TableName("public", "test2"))
        )
        for val in result.values():
            assert_roundtrip_equal(val, df)

    assert not default_log_path.exists()


def test_roundtrip_with_external_hyper_connection(df, tmp_hyper):
    with HyperProcess(Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU) as hyper:
        pantab.frames_to_hyper(
            {"test": df, "test2": df}, tmp_hyper, hyper_process=hyper
        )

        with Connection(hyper.endpoint, tmp_hyper, CreateMode.NONE) as connection:
            result = pantab.frame_from_hyper(connection, table="test")
            assert_roundtrip_equal(result, df)

            result = pantab.frame_from_hyper_query(connection, "SELECT * FROM test")
            assert result.size == 63

            result = pantab.frames_from_hyper(connection)
            assert set(result.keys()) == set(
                (TableName("public", "test"), TableName("public", "test2"))
            )
            for val in result.values():
                assert_roundtrip_equal(val, df)


def test_external_hyper_connection_and_process_error(df, tmp_hyper):
    with HyperProcess(Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU) as hyper:
        with Connection(hyper.endpoint, tmp_hyper, CreateMode.CREATE) as connection:
            expected_msg = (
                "hyper_process parameter is useless because `Connection` is provided"
            )
            with pytest.raises(ValueError, match=expected_msg):
                pantab.frame_from_hyper(connection, table="test", hyper_process=hyper)

            with pytest.raises(ValueError, match=expected_msg):
                pantab.frame_from_hyper_query(
                    connection, "SELECT * FROM test", hyper_process=hyper
                )

            with pytest.raises(ValueError, match=expected_msg):
                pantab.frames_from_hyper(connection, hyper_process=hyper)



__version__ = "2.0.0"

import libpantab  # type: ignore
from ._reader import frame_from_hyper, frame_from_hyper_query, frames_from_hyper
from ._tester import test
from ._writer import frame_to_hyper, frames_to_hyper

__all__ = [
    "__version__",
    "frame_from_hyper",
    "frame_from_hyper_query",
    "frames_from_hyper",
    "frame_to_hyper",
    "frames_to_hyper",
    "test",
]

# We link against HyperAPI in a fun way: In Python, we extract the function
# pointers directly from the Python HyperAPI. We pass those function pointers
# over to the C module which will then use those pointers to directly interact
# with HyperAPI. Furthermore, we check the function signatures to guard
# against API-breaking changes in HyperAPI.
#
# Directly using HyperAPI's C functions always was and still is discouraged and
# unsupported by Tableu. In particular, Tableau will not be able to provide
# official support for this hack.
#
# Because this is highly brittle, we try to make the error message as
# actionable as possible and guide users in the right direction.

api_incompatibility_msg = """
pantab is incompatible with the version of Tableau Hyper API installed on your
system. Please upgrade both `tableauhyperapi` and `pantab` to the latest version.
If doing so does not fix this issue, please file an issue at
https://github.com/innobi/pantab/issues mentioning the exact pantab and HyperAPI
versions which triggered this error.
"""

try:
    from tableauhyperapi.impl.dll import lib, ffi
except ImportError as e:
    raise NotImplementedError(api_incompatibility_msg) from e


def _check_compatibility(check, message):
    if not check:
        raise NotImplementedError(message + "\n" + api_incompatibility_msg)


def _get_hapi_function(name, sig):
    _check_compatibility(hasattr(lib, name), f"function '{name}' missing")
    f = getattr(lib, name)
    func_type = ffi.typeof(f)
    _check_compatibility(
        func_type.kind == "function",
        f"expected '{name}' to be a function, got {func_type.kind}",
    )
    _check_compatibility(
        func_type.cname == sig,
        f"expected '{name}' to have the signature '{sig}', got '{func_type.cname}'",
    )
    return f


libpantab.load_hapi_functions(
    _get_hapi_function("hyper_decode_date", "hyper_date_components_t(*)(uint32_t)"),
    _get_hapi_function("hyper_encode_date", "uint32_t(*)(hyper_date_components_t)"),
    _get_hapi_function("hyper_decode_time", "hyper_time_components_t(*)(uint64_t)"),
    _get_hapi_function("hyper_encode_time", "uint64_t(*)(hyper_time_components_t)"),
    _get_hapi_function(
        "hyper_inserter_buffer_add_null",
        "struct hyper_error_t *(*)(struct hyper_inserter_buffer_t *)",
    ),
    _get_hapi_function(
        "hyper_inserter_buffer_add_bool",
        "struct hyper_error_t *(*)(struct hyper_inserter_buffer_t *, _Bool)",
    ),
    _get_hapi_function(
        "hyper_inserter_buffer_add_int16",
        "struct hyper_error_t *(*)(struct hyper_inserter_buffer_t *, int16_t)",
    ),
    _get_hapi_function(
        "hyper_inserter_buffer_add_int32",
        "struct hyper_error_t *(*)(struct hyper_inserter_buffer_t *, int32_t)",
    ),
    _get_hapi_function(
        "hyper_inserter_buffer_add_int64",
        "struct hyper_error_t *(*)(struct hyper_inserter_buffer_t *, int64_t)",
    ),
    _get_hapi_function(
        "hyper_inserter_buffer_add_double",
        "struct hyper_error_t *(*)(struct hyper_inserter_buffer_t *, double)",
    ),
    _get_hapi_function(
        "hyper_inserter_buffer_add_binary",
        "struct hyper_error_t *(*)(struct hyper_inserter_buffer_t *, uint8_t *, size_t)",
    ),
    _get_hapi_function(
        "hyper_inserter_buffer_add_raw",
        "struct hyper_error_t *(*)(struct hyper_inserter_buffer_t *, uint8_t *, size_t)",
    ),
    _get_hapi_function(
        "hyper_rowset_get_next_chunk",
        "struct hyper_error_t *(*)(struct hyper_rowset_t *, struct hyper_rowset_chunk_t * *)",
    ),
    _get_hapi_function(
        "hyper_destroy_rowset_chunk", "void(*)(struct hyper_rowset_chunk_t *)"
    ),
    _get_hapi_function(
        "hyper_rowset_chunk_field_values",
        "struct hyper_error_t *(*)(struct hyper_rowset_chunk_t *, size_t *, size_t *, uint8_t * * *, size_t * *, int8_t * *)",
    ),
)



from contextlib import contextmanager, nullcontext
from typing import Optional

import tableauhyperapi as tab_api


def ensure_hyper_process(hyper_process: Optional[tab_api.HyperProcess]):
    """
    Spawns an adhoc HyperProcess if needed, i.e. if no existing HyperProcess is provided

    Usage:
    ```
    with ensure_hyper_process(<HyperProcess or None>) as h:
        h.execute_query(...)
    ```
    """
    if hyper_process is None:
        return tab_api.HyperProcess(tab_api.Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU)
    else:
        # Wrap the HyperProcess into a nullcontext such that the `with` doesn't close
        # the HyperProcess
        return nullcontext(hyper_process)


def forbid_hyper_process(hyper_process: Optional[tab_api.HyperProcess]):
    if hyper_process is not None:
        raise ValueError(
            "hyper_process parameter is useless because `Connection` is provided"
        )



import os
import sys
from glob import glob

from setuptools import Extension, find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


# MSVC compiler has different flags; assume that's what we are using on Windows
if os.name == "nt":
    extra_compile_args = ["/WX"]
else:
    extra_compile_args = ["-Wextra", "-Werror"]


pantab_module = Extension(
    "libpantab",
    sources=list(glob("pantab/src/*.c")),
    depends=list(glob("pantab/src/*.h")),
    extra_compile_args=extra_compile_args,
)


setup(
    name="pantab",
    version="2.0.0",
    description="Converts pandas DataFrames into Tableau Hyper Extracts and back",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/WillAyd/pantab",
    author="Will Ayd",
    author_email="william.ayd@icloud.com",
    license="BSD",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Office/Business",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords="tableau visualization pandas dataframe",
    packages=find_packages(),
    package_data={"": ["*.h"], "pantab.tests": ["data/*"]},
    data_files=[("", ["LICENSE.txt", "README.md"])],
    python_requires=">=3.7",
    install_requires=["pandas", "tableauhyperapi"],
    extras_require={"dev": ["pytest"]},
    ext_modules=[pantab_module],
)



import pandas as pd
import pandas.testing as tm
import pytest
from tableauhyperapi import TableName

import pantab
import pantab._compat as compat


def test_read_doesnt_modify_existing_file(df, tmp_hyper):
    pantab.frame_to_hyper(df, tmp_hyper, table="test")
    last_modified = tmp_hyper.stat().st_mtime

    # Try out our read methods
    pantab.frame_from_hyper(tmp_hyper, table="test")
    pantab.frames_from_hyper(tmp_hyper)

    # Neither should not update file stats
    assert last_modified == tmp_hyper.stat().st_mtime


def test_reports_unsupported_type(datapath):
    """
    Test that we report an error if we encounter an unsupported column type.
    Previously, we did not do so but instead assumed that all unsupported columns
    would be string columns. This led to very fascinating failures.
    """
    db_path = datapath / "geography.hyper"
    with pytest.raises(
        TypeError, match=r"Column \"x\" has unsupported datatype GEOGRAPHY"
    ):
        pantab.frame_from_hyper(db_path, table="test")


def test_months_in_interval_raises(df, tmp_hyper, monkeypatch):
    # Monkeypatch a new constructor that hard codes months
    def __init__(self, months: int, days: int, microseconds: int):
        self.months = 1
        self.days = days
        self.microseconds = microseconds

    monkeypatch.setattr(pantab._writer.tab_api.Interval, "__init__", __init__)
    pantab.frame_to_hyper(df, tmp_hyper, table="test")
    with pytest.raises(
        ValueError, match=r"Cannot read Intervals with month components\."
    ):
        pantab.frame_from_hyper(tmp_hyper, table="test")

    with pytest.raises(
        ValueError, match=r"Cannot read Intervals with month components\."
    ):
        pantab.frames_from_hyper(tmp_hyper)


def test_error_on_first_column(df, tmp_hyper, monkeypatch):
    """
    We had a defect due to which pantab segfaulted when an error occured in one of
    the first two columns. This test case is a regression test against that.
    """
    # Monkeypatch a new constructor that hard codes months
    def __init__(self, months: int, days: int, microseconds: int):
        self.months = 1
        self.days = days
        self.microseconds = microseconds

    monkeypatch.setattr(pantab._writer.tab_api.Interval, "__init__", __init__)

    df = pd.DataFrame(
        [[pd.Timedelta("1 days 2 hours 3 minutes 4 seconds")]],
        columns=["timedelta64"],
    ).astype({"timedelta64": "timedelta64[ns]"})
    pantab.frame_to_hyper(df, tmp_hyper, table="test")

    with pytest.raises(
        ValueError, match=r"Cannot read Intervals with month components\."
    ):
        pantab.frame_from_hyper(tmp_hyper, table="test")


def test_read_non_roundtrippable(datapath):
    result = pantab.frame_from_hyper(
        datapath / "dates.hyper", table=TableName("Extract", "Extract")
    )
    expected = pd.DataFrame(
        [["1900-01-01", "2000-01-01"], [pd.NaT, "2050-01-01"]],
        columns=["Date1", "Date2"],
        dtype="datetime64[ns]",
    )
    tm.assert_frame_equal(result, expected)


def test_reads_non_writeable_strings(datapath):
    result = pantab.frame_from_hyper(
        datapath / "non_pantab_writeable.hyper", table=TableName("public", "table")
    )

    expected = pd.DataFrame([["row1"], ["row2"]], columns=["Non-Nullable String"])
    if compat.PANDAS_100:
        expected = expected.astype("string")

    tm.assert_frame_equal(result, expected)


def test_read_query(df, tmp_hyper):
    pantab.frame_to_hyper(df, tmp_hyper, table="test")

    query = "SELECT int16 AS i, '_' || int32 AS _i2 FROM test"
    result = pantab.frame_from_hyper_query(tmp_hyper, query)

    expected = pd.DataFrame(
        [[1, "_2"], [6, "_7"], [0, "_0"]],
        columns=["i", "_i2"],
    )
    str_type = "string" if compat.PANDAS_100 else "object"
    expected = expected.astype({"i": "Int16", "_i2": str_type})

    tm.assert_frame_equal(result, expected)



import pathlib
import shutil
import tempfile
from typing import Dict, Optional, Union

import numpy as np
import pandas as pd
import tableauhyperapi as tab_api

import libpantab  # type: ignore
import pantab._types as pantab_types
from pantab._hyper_util import ensure_hyper_process, forbid_hyper_process

TableType = Union[str, tab_api.Name, tab_api.TableName]


def _read_query_result(
    result: tab_api.Result,
    dtypes: Optional[Dict[str, str]],
) -> pd.DataFrame:
    if dtypes is None:
        dtypes = {}
        # Construct data types from result
        for column in result.schema.columns:
            # `result.schema` does not provide nullability information.
            # Lwt's err on the safe side and always assume they are nullable
            nullability = tab_api.Nullability.NULLABLE
            column_type = pantab_types._ColumnType(column.type, nullability)
            try:
                dtypes[column.name.unescaped] = pantab_types._pandas_types[column_type]
            except KeyError as e:
                raise TypeError(
                    f"Column {column.name} has unsupported datatype {column.type} "
                    f"with nullability {column.nullability}"
                ) from e

    # Call native library to read tuples from result set
    dtype_strs = tuple(dtypes.values())
    df = pd.DataFrame(libpantab.read_hyper_query(result._Result__cdata, dtype_strs))

    df.columns = dtypes.keys()

    # TODO: remove this hackery...
    for k, v in dtypes.items():
        if v == "date":
            dtypes[k] = "datetime64[ns]"

    df = df.astype(dtypes)
    df = df.fillna(value=np.nan)  # Replace any appearances of None

    return df


def _read_table(*, connection: tab_api.Connection, table: TableType) -> pd.DataFrame:
    if isinstance(table, str):
        table = tab_api.TableName(table)

    table_def = connection.catalog.get_table_definition(table)
    columns = table_def.columns

    dtypes: Dict[str, str] = {}
    for column in columns:
        column_type = pantab_types._ColumnType(column.type, column.nullability)
        try:
            dtypes[column.name.unescaped] = pantab_types._pandas_types[column_type]
        except KeyError as e:
            raise TypeError(
                f"Column {column.name} has unsupported datatype {column.type} "
                f"with nullability {column.nullability}"
            ) from e

    query = f"SELECT * from {table}"
    with connection.execute_query(query) as result:
        return _read_query_result(result, dtypes)


def frame_from_hyper(
    source: Union[str, pathlib.Path, tab_api.Connection],
    *,
    table: TableType,
    hyper_process: Optional[tab_api.HyperProcess] = None,
) -> pd.DataFrame:
    """See api.rst for documentation"""

    if isinstance(source, tab_api.Connection):
        forbid_hyper_process(hyper_process)
        return _read_table(connection=source, table=table)
    else:
        with tempfile.TemporaryDirectory() as tmp_dir, ensure_hyper_process(
            hyper_process
        ) as hpe:
            tmp_db = shutil.copy(source, tmp_dir)
            with tab_api.Connection(hpe.endpoint, tmp_db) as connection:
                return _read_table(connection=connection, table=table)


def frames_from_hyper(
    source: Union[str, pathlib.Path, tab_api.Connection],
    *,
    hyper_process: Optional[tab_api.HyperProcess] = None,
) -> Dict[tab_api.TableName, pd.DataFrame]:
    """See api.rst for documentation."""
    result: Dict[TableType, pd.DataFrame] = {}

    if isinstance(source, tab_api.Connection):
        forbid_hyper_process(hyper_process)
        connection = source
        for schema in connection.catalog.get_schema_names():
            for table in connection.catalog.get_table_names(schema=schema):
                result[table] = _read_table(connection=connection, table=table)
    else:
        with tempfile.TemporaryDirectory() as tmp_dir, ensure_hyper_process(
            hyper_process
        ) as hpe:
            tmp_db = shutil.copy(source, tmp_dir)
            with tab_api.Connection(hpe.endpoint, tmp_db) as connection:
                for schema in connection.catalog.get_schema_names():
                    for table in connection.catalog.get_table_names(schema=schema):
                        result[table] = _read_table(connection=connection, table=table)

    return result


def frame_from_hyper_query(
    source: Union[str, pathlib.Path, tab_api.Connection],
    query: str,
    *,
    hyper_process: Optional[tab_api.HyperProcess] = None,
) -> pd.DataFrame:
    """See api.rst for documentation."""

    if isinstance(source, tab_api.Connection):
        forbid_hyper_process(hyper_process)
        with source.execute_query(query) as result:
            return _read_query_result(result, None)
    else:
        with tempfile.TemporaryDirectory() as tmp_dir, ensure_hyper_process(
            hyper_process
        ) as hpe:
            tmp_db = shutil.copy(source, tmp_dir)
            with tab_api.Connection(hpe.endpoint, tmp_db) as connection:
                with connection.execute_query(query) as result:
                    return _read_query_result(result, None)
